//  Copyright (c) 2017-present The blackwidow Authors.  All rights reserved.
//  This source code is licensed under the BSD-style license found in the
//  LICENSE file in the root directory of this source tree. An additional grant
//  of patent rights can be found in the PATENTS file in the same directory.

#include <cmath>
#include <string>
#include <algorithm>
#include "src/redis_hyperloglog.h"
#include "src/blackwidow_murmur3.h"

namespace blackwidow {

const int32_t HLL_HASH_SEED = 313;

HyperLogLog::HyperLogLog(uint8_t precision, std::string origin_register) {
  b_ = precision;
  m_ = 1 << precision;
  alpha_ = Alpha();
  register_ = new char[m_];
  for (uint32_t i = 0; i < m_; ++i) {
    register_[i] = 0;
  }
  if (origin_register != "") {
    for (uint32_t i = 0; i < m_; ++i) {
      register_[i] = origin_register[i];
    }
  }
}

HyperLogLog::~HyperLogLog() {
  delete [] register_;
}

std::string HyperLogLog::Add(const char* value, uint32_t len) {
  uint32_t hash_value;
  MurmurHash3_x86_32(value, len, HLL_HASH_SEED,
                     static_cast<void *>(&hash_value));
  int32_t index = hash_value & ((1 << b_) - 1);
  uint8_t rank = Nclz((hash_value << b_), 32 - b_);
  if (rank > register_[index])
    register_[index] = rank;
  std::string result(m_, 0);
  for (uint32_t i = 0; i < m_; ++i) {
    result[i] = register_[i];
  }
  return result;
}

double HyperLogLog::Estimate() const {
  double estimate = FirstEstimate();
  if (estimate <= 2.5 * m_) {
    uint32_t zeros = CountZero();
    if (zeros != 0) {
      estimate = m_ * log(static_cast<double>(m_) / zeros);
    }
  } else if (estimate > pow(2, 32) / 30.0) {
    estimate = log1p(estimate * -1 / pow(2, 32)) * pow(2, 32) * -1;
  }
  return estimate;
}

double HyperLogLog::FirstEstimate() const {
  double estimate, sum = 0.0;
  for (uint32_t i = 0; i < m_; i++) {
    sum += 1.0 / (1 << register_[i]);
  }

  estimate = alpha_ * m_ * m_ / sum;
  return estimate;
}

double HyperLogLog::Alpha() const {
  switch (m_) {
    case 16:
      return 0.673;
    case 32:
      return 0.697;
    case 64:
      return 0.709;
    default:
      return 0.7213 / (1 + 1.079 / m_);
    }
}

uint32_t HyperLogLog::CountZero() const {
  uint32_t count = 0;
  for (uint32_t i = 0; i < m_; i++) {
    if (register_[i] == 0) {
      count++;
    }
  }
  return count;
}

std::string HyperLogLog::Merge(const HyperLogLog & hll) {
  if (m_ != hll.m_) {
    // TODO(shq) the number of registers doesn't match
  }
  for (uint32_t r = 0; r < m_; r++) {
    if (register_[r] < hll.register_[r]) {
      register_[r] |= hll.register_[r];
    }
  }

  std::string result(m_, 0);
  for (uint32_t i = 0; i < m_; ++i) {
    result[i] = register_[i];
  }
  return result;
}

// ::__builtin_clz(x): 返回左起第一个‘1’之前0的个数
uint8_t HyperLogLog::Nclz(uint32_t x, int b) {
  return (uint8_t)std::min(b, ::__builtin_clz(x)) + 1;
}

}  // namespace blackwidow
